#include <stdio.h>
#include <stdlib.h>
#include "matrix.h"


#define BACKTONE	255
#define OBJTONE		0
#define FILLTONE	10

#define UP   	0
#define DOWN	1
#define E1  	2
#define E2  	3
#define NPARAM 	4

/** Variables for STACK **/

typedef int typexy;
typedef unsigned char uchar;

#define STACK_SIZE 300

typedef struct {
	typexy a, b;
} POS;

POS stack[ STACK_SIZE ];
int pos_stack = -1; 

int cxan=-1,cyan=-1;
/********** Funtiones for the stack ******************/
int push( int a, int b )
{
	int pone;
 
	pone = pos_stack;
	pone++;
	if ( pone >= STACK_SIZE ){
		fprintf( stderr, "Stack is Full %d!\n", pone );
		exit(1);
	}
	stack[pone].a = a;
	stack[pone].b = b;
	pos_stack = pone;
	return( 1 );      /** ok! **/
}

int pop( int *a, int *b )
{
	if ( pos_stack<0 ) return(0);   /** Error **/
	*a = stack[pos_stack].a;
	*b = stack[pos_stack].b;
	pos_stack--;
	return( 1 );    /** Ok! **/
}

int empty_stack( void )
{
	if( pos_stack<0 ) return( 0 ); /** It's empty **/
	return( 1 );                    /** There is something **/
}

void clean_stack( void )
{
	pos_stack = -1;
}

int fill ( MATRIX *IMG, int x, int y )
{
	byte **img;

	img = (byte **)IMG->ptr;
	while ( img[y][x] == OBJTONE ) {
		img[y][x++] = FILLTONE; 
	}
	return(--x);
}

int left ( MATRIX *IMG, int x, int y ) 
{
	byte **img;

	img = (byte **)IMG->ptr;
	while ( img[y][x] == OBJTONE ) {
		--x; 
	}
	return ( ++x );
}

int link( MATRIX *IMG, int x, int y, int *par  )
{
	int above=0, below=0;
	int e1=0, e2 = -1;
	byte **img;

	img = (byte **)IMG->ptr;

	if ( img[y-1][x-1]==OBJTONE ) {
		above++;
		e1 = left( IMG, x-1, y-1 );
		if ( e1<0 ) return( 0 );
	}
	if ( img[y+1][x-1]==OBJTONE ) {
		below++;
		e2 = left( IMG, x-1, y+1 );
		if ( e2<0 ) return( 0 );
	}
	while ( img[y][x] == OBJTONE ) {
		if ( img[y-1][x-1]==BACKTONE && img[y-1][x]==OBJTONE ) {
			if ( above ) push ( e1, y-1 );
			above++;
			e1 = left( IMG, x, y-1 );
			if ( e1<0 ) return( 0 );
		}
		if ( img[y+1][x-1]==BACKTONE && img[y+1][x]==OBJTONE ) {
			if ( below ) push ( e2, y+1 );
			below++;
			e2 = left( IMG, x, y+1 );
			if ( e2<0 ) return( 0 );
		}
		x++;
		if ( x==IMG->cols ) return( 0 );
	}
	if ( img[y-1][x-1]==BACKTONE && img[y-1][x]==OBJTONE ){
		if ( above ) push ( e1, y-1 );
		above++;
		e1 = left( IMG, x, y-1 );
		if ( e1<0 ) return( 0 );
	}
	if ( img[y+1][x-1]==BACKTONE && img[y+1][x]==OBJTONE ){
		if ( below ) push ( e2, y+1 );
		below++;
		e2 = left( IMG, x, y+1 );
		if ( e2<0 ) return( 0 );
	}
	par[UP] = above;
	par[DOWN] = below;
	par[E1] = e1;
	par[E2] = e2;
	return( 1 );
}

int do_fill( MATRIX *IMG, int *val )
{
	int x, y;
	int param[ NPARAM ];
	int pnext, dir;
	int p1x, p1y, p2x, p2y, v;
	int area = 0;
	byte **img;

	p1x = p2x = val[INICIOX];
	p1y = p2y = val[INICIOY];
	img = (byte **)IMG->ptr;
	dir = UP;
	while( pop( &x, &y ) ){
		if ( x < p1x ) p1x = x;
		else if ( x > p2x ) p2x = x;
		if ( y < p1y ) p1y = y;
		else if ( y > p2y ) p2y = y;
		for(;;){
			if ( !link( IMG, x, y, param ) ) return(1);
			if ( !param[UP] && !param[DOWN] ){
				if ( img[y][x]==OBJTONE ) { 
					v = fill( IMG, x, y );
					area += v - x + 1;
					if ( v > p2x ) p2x = v;
				}
				break;
			}

			if ( !param[UP] ) dir=DOWN;
			else if ( !param[DOWN] ) dir=UP;

			if ( param[UP] && param[DOWN] ) {
				if( dir==UP ) push ( param[ E1+DOWN ], y+1 );
				else  push ( param[ E1+UP ], y-1 );
			}
			pnext = param[ E1+dir ];

			v = fill( IMG, x, y );
			area += v - x + 1;
			if ( v > p2x ) p2x = v;

			x = pnext;
			y = (dir==UP) ? y-1 : y+1;

			if ( x < p1x ) p1x = x;
			if ( y < p1y ) p1y = y;
			else if ( y > p2y ) p2y = y;

			if ( y<0 || y==IMG->rows ) return(1);
		}
	}
	val[INICIOX] = p1x;
	val[INICIOY] = p1y;
	val[ANCHO]=p2y-p1y;
	val[LARGO]=p2x-p1x;
	val[AREA] = area;
	return( 0 );
} 

/**
  ----------------------------------------------------------------------
**/


const int tabla[8][2] = {
/** 0 **/ { 0, -1},
/** 1 **/ { 1, -1},
/** 2 **/ { 1,  0},
/** 3 **/ { 1,  1},
/** 4 **/ { 0,  1},
/** 5 **/ {-1,  1},
/** 6 **/ {-1,  0},
/** 7 **/ {-1, -1} };

static int Porden[10];

int checa_punto( MATRIX *IMA, int row, int col )
{
	static int Pt[9];
	int *ptrP, *paux;
	int  i, j, k, sp;
	unsigned char *ptrIMA;

	row--;
	col--;
	k=0;
	for( i=0; i<3; i++ ){
		ptrIMA = (unsigned char *)IMA->ptr[ row+i ];
		ptrIMA += col;
		for( j=0; j<3; j++ ){
			Pt[k] = (*ptrIMA==10) ? 1 : 0;
			k++; ptrIMA++;
		}
	}
	Porden[0] = Pt[4]; Porden[1] = Pt[1]; Porden[2] = Pt[2];
	Porden[3] = Pt[5]; Porden[4] = Pt[8]; Porden[5] = Pt[7];
	Porden[6] = Pt[6]; Porden[7] = Pt[3]; Porden[8] = Pt[0];
	Porden[9] = Pt[1];

	/** No. de transiciones de 0-1 **/
	ptrP = Porden + 1;
	paux = ptrP + 1;
	sp = 0;
	for( i=0; i<8; i++ ) {
		if( *ptrP==0 && *paux>0 ) sp++;
		ptrP++;
		paux++;
	}
	return( sp );
}

void me_voy( int *row, int *col )
{
	int i=-1;

	if ( Porden[1] ) i=0;
	else if ( Porden[3] ) i=2;
	else if ( Porden[5] ) i=4;
	else if ( Porden[7] ) i=6;

	if ( i < 0 ){
		if ( Porden[2] ) i=1;
		else if ( Porden[4] ) i=3;
		else if ( Porden[6] ) i=5;
		else if ( Porden[8] ) i=7;
	}

	*row += tabla[i][1];
	*col += tabla[i][0];
}  

int franja( MATRIX *IN, int row, int col, POINT *pp )
{
	#define FALSE 0
	#define TRUE  1
	int i;
	int flag;
	int cx, cy;
	int sp;
	byte **IMA;
	int termina_franja = FALSE;

	IMA = (byte **)IN->ptr;

	push( row, col );
	flag = 1;
	i = 0;
	while( pop( &cy, &cx ) ) {
		do {
			flag=1;
			if( IMA[cy][cx] == 20 ){
				flag=0;
			}
			sp = checa_punto( IN, cy, cx );
			if ( sp > 1 ) {
				push( cy, cx );
				IMA[cy][cx] = 20;    /** Lo borro **/
				

				if ( flag ) {
					pp[i].x = cx;
					pp[i].y = cy;
					i++;
				}
				me_voy( &cy, &cx );
			}
			else {
				if ( flag ) {
					pp[i].x = cx;
					pp[i].y = cy;
					i++;

					IMA[cy][cx] = 20;    /** Lo borro **/
				}
				if ( !sp )  termina_franja = TRUE;
				else {
					me_voy( &cy, &cx );
				}
			}
			cxan=cx;
			cyan=cy;
		} while( termina_franja == FALSE );
	}
	return ( i );
}
